/*******************************************************************************
**
**  Module: pps.c
**  Date:   12/02/2003
**  Purpose
**  Description: Code file pps.c
**
**                 Copyright 2002, Spectracom Corporation.
**                          ALL RIGHTS RESERVED
**
** Description:
**     The driver that runs the GPS 1-PPS and the Conditioned 1-PPS.  The PPS
**     API described in RFC 2783 is implemented through ioctls.  In this
**     implementation, the only time format supported is PPS_TMFMT_TSPEC
**
********************************************************************************
**
**     6/28/05 Ported from uClinux (2.4.20) to standard Linux (2.4.26) for the
**             HanSolo project.  [MG]
**
**     10/06/09 Ported from HanSolo project into TSync-PCI driver.  [DHS]
**
**     05/18/13 Updated to support new kernel PPSAPI implementation.  [DHS]
**
*******************************************************************************/

#ifdef TSYNC_NTP

#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/pps_kernel.h>

#include "tsync_pps.h"
#include "tsyncpci.h"
#include "tsync_hw_func.h"

/*******************************************************************************
** Defines
*******************************************************************************/
/* #define PPS_DEBUG */
#undef PPS_DEBUG

/*******************************************************************************
** Types
*******************************************************************************/

/*******************************************************************************
** Data Declarations
*******************************************************************************/
static struct pps_device      *pps[TSYNC_PCI_MAX_NUM];
static struct pps_source_info  pps_tsyncpci_info[TSYNC_PCI_MAX_NUM];

/*******************************************************************************
 * Public functions
 ******************************************************************************/

/*******************************************************************************
**
** Function:    pps_init
** Description: Initialize the 1pps devices.
**
** Parameters:
**     IN:      None
**     IN/OUT:  None
**     OUT:     None
**     Returns: None
**
*******************************************************************************/
void pps_init(void)
{
    int i;


    /* Initialize pps structs */
    for (i = 0; i < TSYNC_PCI_MAX_NUM; i++)
    {
    	snprintf(pps_tsyncpci_info[i].name, PPS_MAX_NAME_LEN - 1, "tsyncpci%d", i);
    	snprintf(pps_tsyncpci_info[i].path, PPS_MAX_NAME_LEN - 1, "/dev/pps%d", i);
    	pps_tsyncpci_info[i].mode  = PPS_CAPTUREASSERT  |
    			                     PPS_OFFSETASSERT   |
    			                     PPS_ECHOASSERT     |
    			                     PPS_CANWAIT        |
    			                     PPS_TSFMT_TSPEC;
    	pps_tsyncpci_info[i].owner = THIS_MODULE;
    }

} /* End pps_init */


/*******************************************************************************
**
** Function:    pps_init_one
** Description: Initialize the specified 1pps device specific HW.
**
** Parameters:
**     IN:      tsyncpci_dev_t *hw  - HW device structure
**     IN/OUT:  None
**     OUT:     None
**     Returns: None
**
*******************************************************************************/
int pps_init_one(tsyncpci_dev_t *hw)
{
	unsigned char devNum = hw->slotPosition;


	// Register the PPS source
	pps[devNum] = pps_register_source(&pps_tsyncpci_info[devNum],
			                          PPS_CAPTUREASSERT | PPS_OFFSETASSERT);
	if (pps[devNum] == NULL) {
		pr_err("cannot register PPS source\n");
		return -ENOMEM;
	}

	dev_info(pps[devNum]->dev, "tsyncpci PPS source registered\n");

    /* Unmask the PPS interrupt */
    return(set_hw_int_mask_by_index(hw, 0, 0));

} /* End pps_init_one */


/*******************************************************************************
**
** Function:    pps_remove_one
** Description: Unregister the specified 1pps device specific HW.
**
** Parameters:
**     IN:      tsyncpci_dev_t *hw  - HW device structure
**     IN/OUT:  None
**     OUT:     None
**     Returns: None
**
*******************************************************************************/
void pps_remove_one(tsyncpci_dev_t *hw)
{
	unsigned char devNum = hw->slotPosition;


	dev_info(pps[devNum]->dev, "tsyncpci PPS source unregistered\n");

	// Unregister the PPS source
	pps_unregister_source(pps[devNum]);

} /* End pps_init_one */


/*******************************************************************************
**
** Function:    pps_exit
** Description: The exit function that will be called when this module is
**              unloaded.
**
** Parameters:
**     In:      None
**     In/Out:  None
**     Out:     None
**     Returns: None
**
** Notes:
**    The __exit macro causes the omission of the function when the module is
**    built into the kernel, and like __init, has no effect for
**    loadable modules. Built-in drivers don't need a cleanup function,
**    while loadable modules do.
**
*******************************************************************************/
void pps_exit(void)
{
} /* End pps_exit */


/*******************************************************************************
**
** Function:    pps_isr
** Description: Common interrupt handler for the pps sources.  When an
**              interrupt occurs, the timestamp is updated for this pps's state.
**
** Parameters:
**     In:      None
**     In/Out:  None
**     Out:     None
**     Returns: None
**
*******************************************************************************/
void pps_isr(int src)
{
	struct pps_event_time ts;


    /* Get timestamp */
	pps_get_ts(&ts);

	pps_event(pps[src], &ts, PPS_CAPTUREASSERT, NULL);

} /* End pps_isr */

#endif /* TSYNC_NTP */
