#include "example.h"

#include <time.h>
#include <string.h>
#include <errno.h>

void printUsage(const char *pgmname);

#define OPT_ARG_LIST        "hd:i:vr:f:"

/********************************************************
**
** Test Module: GetTimePerf.c
**
** Purpose:     Get the current time from the board
**
**
*********************************************************/

int main (int argc, char *argv[])
{
    const int      MAX_DEV_NAME_LENGTH = 1024;
    char           deviceName[MAX_DEV_NAME_LENGTH];
    char           *devName = DEFAULT_DEV;
    int            devIdx = DEFAULT_DEV_NUM;

    int            retVal;
    char           *pgmname = argv[0];
    TPRO_BoardObj *pBoard;
    TPRO_TimeObj   tproTime;
    double         *timeArray;
    unsigned char  rv;
    int            loopCntr;

    const int      maxHistogramIndex = 100;
    int            histogram[maxHistogramIndex + 1];

    int             optRc;
    int            verbose = 0;
    int 	   maxRollbacksToPrint = 0;	
    char           *outFileName = NULL;

    const int nQueries = 1000000;
    const int MICROSEC_PER_SEC = 1000000;

    retVal = EXIT_SUCCESS;

    /***************************************************
     * Parse command line
     */
    opterr = 0;
    while ( (optRc = getopt (argc, argv, OPT_ARG_LIST)) != -1 ) {
        switch ( optRc ) {
        case 'd' :
            devName = optarg;
            break;
        case 'i' :
            devIdx = atoi(optarg);
            break;

        case 'v':
            verbose = 1;
            break;
        case 'r':
            maxRollbacksToPrint = atoi(optarg);
            break;
            
        case 'f':
            outFileName = optarg;
            break;
            
        case 'h' :
        default:
            printUsage( pgmname );
            return EXIT_FAILURE;
        }
    }

    if ( optind != argc ) {
        printUsage( pgmname );
        return EXIT_FAILURE;
    }

    
    /***************************************************
     * open device, allocate memory
     */
    snprintf(deviceName, MAX_DEV_NAME_LENGTH, "%s%s%d", DEVICE_ROOT, 
             devName, devIdx);
    
    rv = TPRO_open(&pBoard, deviceName);
    if (rv != TPRO_SUCCESS) {
        fprintf(stderr, "%s: Error opening %s: %s\n", pgmname, deviceName,
                tpro_strerror(rv));
        return EXIT_FAILURE;
    }

    timeArray = (double*)malloc(sizeof(double)*nQueries);
    if (timeArray == NULL) {
        fprintf(stderr, "%s: could not allocate timeArray: %s\n", pgmname,
                strerror(errno));
        return EXIT_FAILURE;
    }

    /**************************************************************
     * Perform a bunch of GetTimes.
     */

    printf("Attempting %u calls to GetTime . . . \n", nQueries);

    for(loopCntr = 0; loopCntr < nQueries; loopCntr++) {
        /* Get the TPRO/TSAT current time */
        rv = TPRO_getTime(pBoard, &tproTime);
        if (rv != TPRO_SUCCESS) {
            break;
        }
        timeArray[loopCntr] = tproTime.secsDouble;
    }
    
    /* If unable to get the TPRO/TSAT current time... */
    if (rv != TPRO_SUCCESS) {
        fprintf(stderr, "%s: Error getting the time from %s: %s\n",
                pgmname, deviceName, tpro_strerror(rv));

        TPRO_close(pBoard);
        free(timeArray);
        return EXIT_FAILURE;
    }

    /***********************************************************
     * Examine the output.
     */
    {

        /* Display the TPRO/TSAT delta time */
        double delta;
        double min = 10.0;
        double max = 0.0;
        double sum = 0.0;

        int rollbackCount = 0;
        int rollbacksPrinted = 0;

        memset(histogram, 0, sizeof(histogram));
        
        for (loopCntr = 0; loopCntr < (nQueries-1); loopCntr++) {
            delta = timeArray[loopCntr+1] - timeArray[loopCntr];
            if (delta < -55.0) {
                delta += 60.0; 
            }
            else if ((delta < 0) && (rollbacksPrinted < maxRollbacksToPrint)) {
                const int surrounding = 5;

                if (loopCntr > surrounding && loopCntr < (nQueries-surrounding)) {
                    printf("Rollback detected:\n");
                    int iQuery;
                    for (iQuery = loopCntr - surrounding; iQuery <= loopCntr + surrounding; iQuery++) {
                        printf("  %8u: %f %s\n", iQuery, timeArray[iQuery],
                               (iQuery == (loopCntr+1))? "*":" ");
                    }
                    printf("\n");
                    rollbacksPrinted++;
                }
                rollbackCount++;
            }

            if (delta < min) {
                min = delta;
            }

            if (delta > max) {
                max = delta;
            }

            sum += delta;

            unsigned int deltaMicrosecs = delta * MICROSEC_PER_SEC;
            if (deltaMicrosecs < maxHistogramIndex) {
                histogram[deltaMicrosecs]++;
            }
            else {
                histogram[maxHistogramIndex]++;
            }
        }

        // Account for the fact the sometimes the first element is
        // larger than the last element, due to rollover.
        delta = timeArray[nQueries - 1] - timeArray[0];

        if (delta < -50.0) {
            delta += 60.0;
        }

        if (rollbackCount > 0) { printf (" Rollbacks    : %u\n", rollbackCount); }
        printf (" Total time   : %.6f seconds\n", delta);
        printf (" Ave interval : %.7f seconds\n", sum/(double)nQueries);
        printf (" Rate         : %.1f Hz\n", (double)nQueries / delta);
        printf (" Max interval : %.6f seconds\n", max);
        printf (" Min interval : %.6f seconds\n", min);

        if (verbose > 0) {
            unsigned int histoSum = 0;
            for (loopCntr = 0; loopCntr < maxHistogramIndex; loopCntr++) {
                if (histogram[loopCntr] > 0) {
                    histoSum += histogram[loopCntr];
                    printf("  %3u microseconds apart: %8u (%4.1f%%)\n",
                           loopCntr,
                           histogram[loopCntr],
                           (histogram[loopCntr]/(double)nQueries)*100.0);
                }
            }
            
            if (histogram[maxHistogramIndex] > 0) {
                histoSum += histogram[maxHistogramIndex];
                printf(">=%3u microseconds apart: %8u (%4.1f%%)\n",
                       maxHistogramIndex,
                       histogram[maxHistogramIndex],
                       (histogram[maxHistogramIndex]/(double)nQueries)*100.0);
            }
        }
    }

    if (outFileName != NULL) {
        FILE *outFile;

        int iQuery;

        outFile = fopen(outFileName, "w");
        if (outFile != NULL) {
            printf("Writing query results to file %s . . . \n", outFileName);

            for (iQuery = 0; iQuery < nQueries; iQuery++) {
                char *notes = " ";
                if ((iQuery > 0) && (timeArray[iQuery] < timeArray[iQuery-1])) {
                    notes = "Rollback.";
                    if ((timeArray[iQuery] < 1) && (timeArray[iQuery-1] > 59)) {
                        notes = "minute rollover.";
                    }
                }
                fprintf(outFile, "%8u: %9.6f %s\n", iQuery, timeArray[iQuery], notes);
            }
            
            fclose(outFile);
            printf("done.\n");
        }
        else {
            fprintf(stderr, "%s: unable to open %s for writing: %s\n",
                    pgmname, outFileName, strerror(errno));
            retVal = EXIT_FAILURE;
        }
    }

    free(timeArray);

    /* Close the TPRO/TSAT device */
    rv = TPRO_close(pBoard);
    if (rv != TPRO_SUCCESS) {
        fprintf (stderr, "%s: Unable to close %s: %s\n", pgmname, deviceName, tpro_strerror(rv));
        retVal = EXIT_FAILURE;
    }

    return (retVal);
} /* End - main() */

void printUsage(const char *pgmname) {
    (void)printf("Usage: %s [-h] [-d NAME] [-i INDEX] [-r rollbacks] [-f file]\n\n", pgmname );
    (void)printf("\t-h   help message\n");
    (void)printf("\n");
    (void)printf("\t-d   device name  (default: %s)\n", DEFAULT_DEV);
    (void)printf("\t-i   device index (default: %d)\n", DEFAULT_DEV_NUM);
    (void)printf("\n");
    (void)printf("\t-v   verbose output (query delta tallies)\n");
    (void)printf("\t-r   maximum rollbacks to print in detail (default: none)\n");
    (void)printf("\t-f   write all query values to the named file\n");
    
}    
